package space.linuxct.keyrollbackdemo

import android.content.Context
import android.content.pm.PackageInfo
import android.content.pm.PackageManager
import android.os.Bundle
import android.security.keystore.KeyGenParameterSpec
import android.security.keystore.KeyProperties
import android.util.Base64
import androidx.activity.ComponentActivity
import androidx.activity.compose.setContent
import androidx.compose.foundation.layout.Box
import androidx.compose.foundation.layout.fillMaxSize
import androidx.compose.material3.MaterialTheme
import androidx.compose.material3.Surface
import androidx.compose.material3.Text
import androidx.compose.runtime.Composable
import androidx.compose.ui.Alignment
import androidx.compose.ui.Modifier
import androidx.compose.ui.tooling.preview.Preview
import space.linuxct.keyrollbackdemo.ui.theme.KeyRollbackDemoTheme
import java.security.KeyPairGenerator
import java.security.KeyStore
import java.util.*


class MainActivity : ComponentActivity() {
    override fun onCreate(savedInstanceState: Bundle?) {
        super.onCreate(savedInstanceState)

        val pInfo: PackageInfo = applicationContext.packageManager.getPackageInfo(packageName, 0)
        val version = pInfo.longVersionCode

        val sp = applicationContext.getSharedPreferences("KeyRollbackDemoApp", Context.MODE_PRIVATE)
        val previousVersionCode = sp.getLong("VersionCode", -1L)
        sp.edit().putLong("VersionCode", version).apply()


        if (version == 1L){
            if (previousVersionCode == 2L) {
                // If this happens, it means the SharedPreferences were not rolled back correctly, so something went wrong.
                deleteKeyPair()
            }

            if (!doesKeyExist()){
                // Should be called mainly when the app is first started, other scenarios should not happen
                generateKeyPair()
            }
        }

        if (version == 2L){
            if (previousVersionCode == 1L){
                // Application upgraded, so we delete and regenerate the keypair
                deleteKeyPair()
                generateKeyPair()
            }
        }

        val string = getPublicKey()

        setContent {
            KeyRollbackDemoTheme {
                // A surface container using the 'background' color from the theme
                Surface(
                    modifier = Modifier.fillMaxSize(),
                    color = MaterialTheme.colorScheme.background,
                ) {
                    Box(
                        contentAlignment = Alignment.Center,
                        modifier = Modifier.fillMaxSize()
                    ) {
                        DisplayContents(string, version, previousVersionCode)
                    }
                }
            }
        }
    }

    companion object {
        private const val ORIGINATION_TIME_OFFSET = 1000000
        private const val CONSUMPTION_TIME_OFFSET = 2000000
    }

    private fun doesKeyExist(): Boolean {
        val ks = KeyStore.getInstance("AndroidKeyStore")
        ks.load(null)
        return ks.containsAlias("testAlias")
    }

    private fun generateKeyPair(){
        val hasStrongBox = try { applicationContext.packageManager.hasSystemFeature(PackageManager.FEATURE_STRONGBOX_KEYSTORE) }
            catch (e: Throwable) { false }

        val keyPairGenerator = KeyPairGenerator.getInstance(KeyProperties.KEY_ALGORITHM_RSA, "AndroidKeyStore")
        val now = Date()
        val originationEnd = Date(now.time + ORIGINATION_TIME_OFFSET)
        val consumptionEnd = Date(now.time + CONSUMPTION_TIME_OFFSET)
        val builder =
            KeyGenParameterSpec.Builder("testAlias", KeyProperties.PURPOSE_SIGN)
                .setKeySize(512)
                .setKeyValidityStart(now)
                .setKeyValidityForOriginationEnd(originationEnd)
                .setKeyValidityForConsumptionEnd(consumptionEnd)
                .setAttestationChallenge("testAttestationChallenge".toByteArray())
        if (hasStrongBox) {
            builder.setDigests(KeyProperties.DIGEST_NONE, KeyProperties.DIGEST_SHA256)
            builder.setIsStrongBoxBacked(true)
        } else {
            builder.setDigests(KeyProperties.DIGEST_NONE, KeyProperties.DIGEST_SHA256, KeyProperties.DIGEST_SHA512)
        }
        keyPairGenerator.initialize(builder.build())
        keyPairGenerator.generateKeyPair()
    }

    private fun deleteKeyPair(){
        val ks = KeyStore.getInstance("AndroidKeyStore")
        ks.load(null)
        ks.deleteEntry("testAlias")
    }

    private fun getPublicKey(): String{
        val ks = KeyStore.getInstance("AndroidKeyStore")
        ks.load(null)
        val pk = ks.getCertificate("testAlias").publicKey.encoded
        return Base64.encodeToString(pk, Base64.DEFAULT)
    }
}

@Composable
fun DisplayContents(pk: String, versionCode: Long, previousVersionCode: Long) {
    Text(text = "Public Key (Base64-encoded): $pk\nVersion code: $versionCode\n\nPrevious version code: $previousVersionCode")
}

@Preview(showBackground = true)
@Composable
fun DefaultPreview() {
    KeyRollbackDemoTheme {
        DisplayContents("AAAAAAAAAAAAAAA==", 1L, -1L)
    }
}